package main

import (
	"fmt"
	"strings"
	"sync"
	"sync/atomic"

	"huawei.com/goclient/huaweichain/bstore"
	"huawei.com/goclient/utils"
	"huawei.com/huaweichain/sdk/client"
)

func main() {
	gatewayClient, err := client.NewGatewayClient(utils.AppConfig().ConfigFilePath)
	if err != nil {
		fmt.Println("init new gateway client error")
	}

	endorseNodes := strings.Split(utils.AppConfig().EndorserNodes, ",")
	net, err := utils.NewNodes(gatewayClient, endorseNodes, utils.AppConfig().ConsensusNode)
	if err != nil {
		fmt.Printf("new nodes network error: %v\n", err)
		return
	}

	txEvent, err := net.EventListener.EventAction.GetTxEventService(utils.AppConfig().ChainID)
	if err != nil {
		fmt.Printf("event action register tx event error: %v\n", err)
		return
	}
	defer txEvent.Close()

	utils.SetTxEvent(txEvent)

	//1. function usage example in contract.go(合约)
	var wg sync.WaitGroup
	var txHashID string
	var count int32 = 0
	for i := 0; i < 5; i++ {
		wg.Add(1)
		go func(count int32) {
			defer wg.Done()
			insertName := fmt.Sprintf("%d;huaweichain_%d", count, count)
			_, txHashID, err = utils.Send(gatewayClient, net, utils.AppConfig(), "insert", insertName)
			fmt.Println(count, txHashID)
			if err != nil {
				fmt.Println(err)
			}
		}(count)
		atomic.AddInt32(&count, 1)
	}
	wg.Wait()

	_, err = utils.Query(gatewayClient, net, utils.AppConfig(), "query", "1")
	if err != nil {
		fmt.Println(err)
	}

	value, err := utils.Query(gatewayClient, net, utils.AppConfig(), "queryInRange", "1;10") // key区间为[a, c)
	if err != nil {
		fmt.Println(err)
	}
	fmt.Println("range query result is ", value)

	value, err = utils.Query(gatewayClient, net, utils.AppConfig(), "queryInRangeMatchCondition", "1;10;!=;huaweichain_5") // 查询区间为[a,c)、判断符为!=、比较值为001
	if err != nil {
		fmt.Println(err)
	}
	fmt.Println("condition query in range result is ", value)

	// 2. function usage example in tx.go(交易)
	txTools := utils.TxTool{}
	tx, err := txTools.QueryTxByTxID(gatewayClient, utils.AppConfig(), txHashID)
	if err != nil {
		fmt.Println(err)
	}

	txResult, err := txTools.QueryTxResultByTxID(*gatewayClient, utils.AppConfig(), txHashID)
	if err != nil || txResult == nil {
		fmt.Println(err)
	}
	fmt.Println("\ntx`s result is ", txResult.Status)
	txTools.PrintTxInfo(*tx)

	// 3. function usage example in block.go(区块)
	blockTool := utils.BlockTool{}
	block, err := blockTool.QueryBlockByNumber(gatewayClient, utils.AppConfig(), "3")
	if err != nil {
		fmt.Println(err)
	}

	block, err = blockTool.QueryBlockByTxID(gatewayClient, utils.AppConfig(), txHashID)
	if err != nil {
		fmt.Println(err)
	}

	blockNumber, err := blockTool.QueryLastBlockNumber(gatewayClient, utils.AppConfig())
	if err != nil {
		fmt.Println(err)
	}
	fmt.Println("block`s last number is ", blockNumber)
	blockTool.PrintBlockBasicInfo(*block)
	blockTool.PrintBlockTxsInfo(block)

	// 4. rich media storage example
	if endorseNodes == nil || len(endorseNodes) < 1 || endorseNodes[0] == "" {
		fmt.Println("Failed to init storage client, error: endorser node not specified.")
		return
	}
	bsClient, err := bstore.NewBsClient(gatewayClient, utils.AppConfig().ChainID, endorseNodes[0], utils.AppConfig().ConsensusNode)
	if err != nil {
		fmt.Println(err)
		return
	}

	// upload file to chain
	fileInfo, txHashID, err := bsClient.UploadFile("D:/goDemo/example.png", "example.png")
	if err != nil {
		fmt.Println(err)
	} else {
		fmt.Println(fmt.Sprintf("upload file to chain success! file name: %s, file hash: %s, version: %s, txID: %s",
			fileInfo.FileName, fileInfo.FileHash, fileInfo.Version, txHashID))
	}

	// download file from chain
	err = bsClient.DownloadFile("D:/goDemo/exampleDownload.png", "example.png", 1)
	if err != nil {
		fmt.Println(err)
	} else {
		fmt.Println("download file from chain success!")
	}

	// query file history on chain
	fileHistories, err := bsClient.GetFileHistory("example.png")
	if err != nil {
		fmt.Println(err)
	} else {
		fmt.Println("get file history success!")
		for _, history := range fileHistories {
			fmt.Println(fmt.Sprintf("create time: %s, uploader: %s, updated time: %s, hashcode: %s, version: %d",
				history.CreatedTime, history.Uploader, history.UpdatedTime, history.HashCode, history.Version))
		}
	}

	// query file operation records on chain
	fileOperations, err := bsClient.GetFileOperation("example.png", "1645440974728", "1645441080861")
	if err != nil {
		fmt.Println(err)
	} else {
		fmt.Println("get file operation record success!")
		for _, operation := range fileOperations {
			fmt.Println(fmt.Sprintf("operator: %s, time: %s, event type: %s", operation.Operator, operation.Time, operation.EventType))
		}
	}
}
